/*****************************************************************************

        ParamEq.cpp
        Copyright (c) 2002-2006 Laurent de Soras

--- Legal stuff ---

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130) // "'operator' : logical operation on address of string constant"
	#pragma warning (1 : 4223) // "nonstandard extension used : non-lvalue array converted to pointer"
	#pragma warning (1 : 4705) // "statement has no effect"
	#pragma warning (1 : 4706) // "assignment within conditional expression"
	#pragma warning (4 : 4786) // "identifier was truncated to '255' characters in the debug information"
	#pragma warning (4 : 4800) // "forcing value to bool 'true' or 'false' (performance warning)"
	#pragma warning (4 : 4355) // "'this' : used in base member initializer list"
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"basic/basic_def.h"
#include	"basic/basic_fnc.h"
#include	"dsp/dsp_def.h"
#include	"ParamEq.h"

#include	<cassert>
#include	<cmath>
#include	<cstdio>



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



ParamEq::ParamEq ()
:	_nbr_bands (0)
,	_sample_freq (0)
{
	assert (NBR_GLOB_PARAM >= 0);
	assert (NBR_TRACK_PARAM >= 0);

	assert (MAX_PITCH < 0xFFFF);
	assert (MAX_GAIN < 0xFF);
	assert (MAX_Q < 0xFF);

	GlobalVals = &_public_param_pack._gval;
	TrackVals = &_public_param_pack._tval;
}



::CMDKMachineInterfaceEx *	ParamEq::GetEx ()
{
	return (&_dummy_miex);
}



void	ParamEq::OutputModeChanged (bool stereo)
{
	// Nothing
}



bool	ParamEq::MDKWork (float *psamples, int numsamples, int const mode)
{
	if (! (mode & WM_READ) || ! (mode & WM_WRITE))
	{
		return (false);
	}

	psamples [0              ] += 7 * dsp::ANTI_DENORMAL;
	psamples [numsamples >> 1] -= 3 * dsp::ANTI_DENORMAL;

	float *		buf_ptr_arr [EqBand::NBR_CHN] =
	{
		psamples,
		psamples
	};

	for (int band = 0; band < _nbr_bands; ++band)
	{
		_band_arr [band].process (buf_ptr_arr, numsamples, 1);
	}

	return (true);
}



bool	ParamEq::MDKWorkStereo (float *psamples, int numsamples, int const mode)
{
	if (! (mode & WM_READ) || ! (mode & WM_WRITE))
	{
		return (false);
	}

	float *		buf_ptr_arr [EqBand::NBR_CHN] =
	{
		&_buffer_arr [0] [0],
		&_buffer_arr [1] [0]
	};

	copy_2i_2 (buf_ptr_arr [0], buf_ptr_arr [1], psamples, numsamples);

	for (int band = 0; band < _nbr_bands; ++band)
	{
		_band_arr [band].process (buf_ptr_arr, numsamples, 2);
	}

	copy_2_2i (psamples, buf_ptr_arr [0], buf_ptr_arr [1], numsamples);

	return (true);
}



void	ParamEq::MDKInit (::CMachineDataInput * const pi)
{
	// Misc DSP init
	{
		_sample_freq = float (pMasterInfo->SamplesPerSec);
		assert (_sample_freq > 0);
		for (int track = 0; track < MAX_NBR_BANDS; ++track)
		{
			_band_arr [track].set_sample_freq (_sample_freq);
		}
	}

	// Set default settings
	{
		_nbr_bands = MIN_NBR_BANDS;
		for (int track = 0; track < MAX_NBR_BANDS; ++track)
		{
			set_default_track_settings (track);
		}
	}

	// Overwrite default settings with preset data if any
	if (pi != 0)
	{
		byte			version;
		byte			nbr_tracks;

		pi->Read (version);

		pi->Read (nbr_tracks);
		_nbr_bands = basic::limit (
			int (nbr_tracks), int (MIN_NBR_BANDS), int (MAX_NBR_BANDS)
		);

		pi->Read (&_public_param_pack, sizeof (_public_param_pack));
	}

	// Apply settings
	{
		for (int track = 0; track < MAX_NBR_BANDS; ++track)
		{
			apply_track_settings (track);
		}
	}
}



void	ParamEq::MDKSave (::CMachineDataOutput * const po)
{
	assert (po != 0);

	po->Write (static_cast <byte> (MCH_VERSION));

	po->Write (static_cast <byte> (_nbr_bands));
	po->Write (&_savable_param_pack, sizeof (_savable_param_pack));
}



void	ParamEq::Tick ()
{
	for (int track = 0; track < _nbr_bands; ++track)
	{
		apply_track_settings (track);
	}
}



void	ParamEq::SetNumTracks (int const n)
{
	for (int track = _nbr_bands; track < n; ++track)
	{
		set_default_track_settings (track);
		apply_track_settings (track);
		_band_arr [track].clear_buffers ();
	}

	_nbr_bands = n;
}



const char *	ParamEq::DescribeValue (int const param, int const value)
{
	const long		max_len = 63;
	static char		txt_0 [max_len + 1];

	switch (param)
	{
	case	ParamIndex_TYPE:
		sprintf (txt_0, "%s", EqBand::get_type_name (buzz_to_type (value)));
		break;

	case	ParamIndex_FREQ:
		sprintf (txt_0, "%d Hz", basic::round_int (buzz_to_freq (value)));
		break;

	case	ParamIndex_GAIN:
		sprintf (txt_0, "%+.1f dB", log (buzz_to_gain (value)) * (20 / basic::LN10));
		break;

	case	ParamIndex_Q:
		sprintf (txt_0, "%2.2f", double (buzz_to_q (value)));
		break;

	default:
		txt_0 [0] = '\0';
		break;
	}

	return (txt_0);
}



const ::CMachineInfo	ParamEq::_mac_info =
{
	MT_EFFECT,								// type, MT_GENERATOR or MT_EFFECT
	MI_VERSION,								// version, always set this to MI_VERSION
	MIF_DOES_INPUT_MIXING,				// Flags ????
	MIN_NBR_BANDS,							// minimum number of tracks
	MAX_NBR_BANDS,							// maximum number of tracks
	NBR_GLOB_PARAM,						// number of global parameters
	NBR_TRACK_PARAM,						// number of track parameters
	_param_list,							// pointer to list of parameters
	0,											// number of attributes
	0,											// pointer to list of attributes
#if ! defined (NDEBUG)
	"FireSledge ParamEQ (debug build)",	// name for debug build
#else
	"FireSledge ParamEQ",				// name for release build
#endif
	"ParamEQ",								// short name
	"FireSledge",							// your name
	0											// list of commands
};



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



EqBand::Type	ParamEq::buzz_to_type (int param) const
{
	return (EqBand::Type (param));
}



float	ParamEq::buzz_to_freq (int param) const
{
	const float		freq = static_cast <float> (
		MIN_FREQ * exp (param * (basic::LN2 / PITCH_RESOL))
	);

	return (freq);
}



float	ParamEq::buzz_to_gain (int param) const
{
	const int		scale = GAIN_AMP_L2 * GAIN_RESOL;
	const int		unbiased = param - scale;
	const float		shaped = reshape (unbiased, scale);
	const float		gain = static_cast <float> (
		exp (shaped * (basic::LN2 * GAIN_AMP_L2))
	);

	return (gain);
}



float	ParamEq::buzz_to_q (int param) const
{
	const int		scale = Q_AMP_L2 * Q_RESOL;
	const int		unbiased = param - scale;
	const float		shaped = reshape (unbiased, scale);
	const float		q = static_cast <float> (
		exp (shaped * (basic::LN2 * Q_AMP_L2))
	);

	return (q);
}



void	ParamEq::set_type (int band, EqBand::Type type)
{
	_band_arr [band].set_parameters (
		type,
		_band_arr [band].get_freq (),
		_band_arr [band].get_gain (),
		_band_arr [band].get_q ()
	);
}



void	ParamEq::set_freq (int band, float freq)
{
	_band_arr [band].set_parameters (
		_band_arr [band].get_type (),
		freq,
		_band_arr [band].get_gain (),
		_band_arr [band].get_q ()
	);
}



void	ParamEq::set_gain (int band, float gain)
{
	_band_arr [band].set_parameters (
		_band_arr [band].get_type (),
		_band_arr [band].get_freq (),
		gain,
		_band_arr [band].get_q ()
	);
}



void	ParamEq::set_q (int band, float q)
{
	_band_arr [band].set_parameters (
		_band_arr [band].get_type (),
		_band_arr [band].get_freq (),
		_band_arr [band].get_gain (),
		q
	);
}



void	ParamEq::set_default_track_settings (int track)
{
	assert (track >= 0);
	assert (track < MAX_NBR_BANDS);

	TVals &			tval = _public_param_pack._tval [track];

	tval._type = _param_type.DefValue;
	tval._freq = _param_freq.DefValue;
	tval._gain = _param_gain.DefValue;
	tval._q    = _param_q.DefValue;
}



void	ParamEq::apply_track_settings (int track)
{
	assert (track >= 0);
	assert (track < MAX_NBR_BANDS);

	const TVals &	tval = _public_param_pack._tval [track];
	TVals &			stval = _savable_param_pack._tval [track];

	if (tval._type != _param_type.NoValue)
	{
		set_type (track, buzz_to_type (tval._type));
		stval._type = tval._type;
	}

	if (tval._freq != _param_freq.NoValue)
	{
		set_freq (track, buzz_to_freq (tval._freq));
		stval._freq = tval._freq;
	}

	if (tval._gain != _param_gain.NoValue)
	{
		set_gain (track, buzz_to_gain (tval._gain));
		stval._gain = tval._gain;
	}
	
	if (tval._q != _param_q.NoValue)
	{
		set_q (track, buzz_to_q (tval._q));
		stval._q = tval._q;
	}
}



void	ParamEq::copy_2_2i (float out_ptr [], const float in_1_ptr [], const float in_2_ptr [], long nbr_spl)
{
	assert (out_ptr != 0);
	assert (in_1_ptr != 0);
	assert (in_2_ptr != 0);
	assert (nbr_spl > 0);

	long				pos = 0;
	do
	{
		out_ptr [pos * 2    ] = in_1_ptr [pos];
		out_ptr [pos * 2 + 1] = in_2_ptr [pos];
		++ pos;
	}
	while (pos < nbr_spl);
}



void	ParamEq::copy_2i_2 (float out_1_ptr [], float out_2_ptr [], const float in_ptr [], long nbr_spl)
{
	assert (out_1_ptr != 0);
	assert (out_2_ptr != 0);
	assert (in_ptr != 0);
	assert (nbr_spl > 0);

	long				pos = 0;
	do
	{
		out_1_ptr [pos] = in_ptr [pos * 2    ];
		out_2_ptr [pos] = in_ptr [pos * 2 + 1];
		++ pos;
	}
	while (pos < nbr_spl);
}



float	ParamEq::reshape (int value, int scale)
{
	assert (scale > 0);
	assert (value >= -scale);
	assert (value <= scale);

	const float		norm_val = float (value) / scale;
	const float		weight = 3;
	const float		mult = 1 / (1 + weight);
	const float		shaped = static_cast <float> (
		norm_val * (1 + fabs (norm_val) * weight) * mult
	);

	return (shaped);
}



const ::CMachineParameter *	ParamEq::_param_list [NBR_GLOB_PARAM + NBR_TRACK_PARAM] =
{
	// Global param

	// Track param
	&_param_type,
	&_param_freq,
	&_param_gain,
	&_param_q
};



const ::CMachineParameter	ParamEq::_param_type =
{
	pt_byte,								// type
	"FilterType",						// short (one or two words) name 
	"Filter Type (0 = peak, 1 = low shelf, 2 = high shelf)",	// longer description 
	0x00,									// minumum value
	EqBand::Type_NBR_ELT - 1,		// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	EqBand::Type_PEAK					// Default value
};

const ::CMachineParameter	ParamEq::_param_freq =
{
	pt_word,								// type
	"Frequency",						// short (one or two words) name 
	"Frequency (log scale)",		// longer description 
	0x0000,								// minumum value
	MAX_PITCH,							// maximum value
	0xFFFF,								// some value outside [min..max] range 
	MPF_STATE,							// flags
	MAX_PITCH / 2						// Default value
};

const ::CMachineParameter	ParamEq::_param_gain =
{
	pt_byte,								// type
	"Gain",								// short (one or two words) name 
	"Gain (dB)",						// longer description 
	0x00,									// minumum value
	MAX_GAIN,							// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	MAX_GAIN / 2						// Default value
};

const ::CMachineParameter	ParamEq::_param_q =
{
	pt_byte,								// type
	"Q",									// short (one or two words) name 
	"Q (log scale)",					// longer description 
	0x00,									// minumum value
	MAX_Q,								// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	MAX_Q / 2							// Default value
};



const char	ParamEq::COPYRIGHT_0 [] =
	"@(#) ParamEQ / Build " __DATE__ " / Copyright (c) FireSledge / Freeware.\n"
	"Laurent de Soras\n"
	"laurent.de.soras(at)club-internet.fr\n"
	"4 avenue Alsace-Lorraine\n"
	"92500 Rueil-Malmaison\n"
	"France";



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/