/*****************************************************************************

        ParamEq.h
        Copyright (c) 2002-2006 Laurent de Soras

--- Legal stuff ---

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*Tab=3***********************************************************************/



#if ! defined (ParamEq_HEADER_INCLUDED)
#define	ParamEq_HEADER_INCLUDED

#if defined (_MSC_VER)
	#pragma once
	#pragma warning (4 : 4250) // "Inherits via dominance."
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"EqBand.h"
#include	"mdk.h"



class ParamEq
:	public ::CMDKMachineInterface
{

/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

public:

	enum {			MCH_VERSION			= 1		};
	enum {			MIN_NBR_BANDS		= 4		};
	enum {			MAX_NBR_BANDS		= 16		};

						ParamEq ();
	virtual			~ParamEq () {}

	// CMDKMachineInterface
	virtual ::CMDKMachineInterfaceEx *
						GetEx ();
	virtual void	OutputModeChanged (bool stereo);
	virtual bool	MDKWork (float *psamples, int numsamples, int const mode);
	virtual bool	MDKWorkStereo (float *psamples, int numsamples, int const mode);
	virtual void	MDKInit (::CMachineDataInput * const pi);
	virtual void	MDKSave (::CMachineDataOutput * const po);

	// CMachineInterface
	virtual void	Tick ();
	virtual void	SetNumTracks (int const n);
	virtual const char *
						DescribeValue (int const param, int const value);

	// Buzz stuff
	static const ::CMachineInfo
						_mac_info;



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

protected:



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

private:

	enum {			PITCH_RANGE_L2		= 10		};	// Power of 2 of the ratio between min and max frequencies
	enum {			PITCH_RESOL			= 1024	};	// Subdivisions of each power of 2 for the pitch
	enum {			MAX_PITCH			= PITCH_RANGE_L2 * PITCH_RESOL	};
	enum {			MIN_FREQ				= 20		};	// Hz
	enum {			MAX_FREQ				= MIN_FREQ << PITCH_RANGE_L2	};
	enum {			Q_AMP_L2				= 4		};	// +/-
	enum {			Q_RESOL				= 16		};
	enum {			MAX_Q					= Q_AMP_L2 * Q_RESOL * 2	};
	enum {			GAIN_AMP_L2			= 6		};
	enum {			GAIN_RESOL			= 16		};
	enum {			MAX_GAIN				= GAIN_AMP_L2 * GAIN_RESOL * 2	};

	enum ParamIndex
	{
		ParamIndex_TRACK_PARAM_ZONE = 0,
		ParamIndex_TYPE = ParamIndex_TRACK_PARAM_ZONE,
		ParamIndex_FREQ,
		ParamIndex_GAIN,
		ParamIndex_Q,

		ParamIndex_NBR_PARAM
	};

	enum {			NBR_GLOB_PARAM		= ParamIndex_TRACK_PARAM_ZONE	};
	enum {			NBR_TRACK_PARAM	= ParamIndex_NBR_PARAM - ParamIndex_TRACK_PARAM_ZONE	};

#pragma pack (push, 1)	// <- don't forget that

	class GVals
	{
	public:
	};

	class TVals
	{
	public:
		byte				_type;
		word				_freq;	// Hz, ]0 ; fs/2[
		byte				_gain;	// Linear, >= 0
		byte				_q;		// Linear, > 0
	};

	class ParamPack
	{
	public:
		GVals				_gval;
		TVals				_tval [MAX_NBR_BANDS];
	};

#pragma pack (pop)

	class DummyMiEx
	:	public ::CMDKMachineInterfaceEx
	{
	};

	EqBand::Type	buzz_to_type (int param) const;
	float				buzz_to_freq (int param) const;
	float				buzz_to_gain (int param) const;
	float				buzz_to_q (int param) const;

	void				set_type (int band, EqBand::Type type);
	void				set_freq (int band, float freq);
	void				set_gain (int band, float gain);
	void				set_q (int band, float q);

	void				set_default_track_settings (int track);
	void				apply_track_settings (int track);

	static void		copy_2_2i (float out_ptr [], const float in_1_ptr [], const float in_2_ptr [], long nbr_spl);
	static void		copy_2i_2 (float out_1_ptr [], float out_2_ptr [], const float in_ptr [], long nbr_spl);
	static float	reshape (int value, int scale);

	int				_nbr_bands;
	EqBand			_band_arr [MAX_NBR_BANDS];
	float				_buffer_arr [EqBand::NBR_CHN] [MAX_BUFFER_LENGTH];
	float				_sample_freq;		// Sample frequency, Hz, > 0

	// Buzz stuff
	ParamPack		_public_param_pack;
	ParamPack		_savable_param_pack;
	DummyMiEx		_dummy_miex;

	static const ::CMachineParameter *
						_param_list [NBR_GLOB_PARAM + NBR_TRACK_PARAM];

	static const ::CMachineParameter
						_param_type;
	static const ::CMachineParameter
						_param_freq;
	static const ::CMachineParameter
						_param_q;
	static const ::CMachineParameter
						_param_gain;

	static const char
						COPYRIGHT_0 [];



/*\\\ FORBIDDEN MEMBER FUNCTIONS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

private:

						ParamEq (const ParamEq &other);
	ParamEq &		operator = (const ParamEq &other);
	bool				operator == (const ParamEq &other);
	bool				operator != (const ParamEq &other);

};	// class ParamEq



//#include	"ParamEq.hpp"



#endif	// ParamEq_HEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
